// raytracing special relativity
// by jph

#define FRAMES    24
#define RECURSION 4
#define THRESHOLD .02

#define MATHEMATICA_V 0
#define MATHEMATICA_X 1
#define MATHEMATICA_Y 2
#define MATHEMATICA_Z 3
#define MATHEMATICA_T 4

#define LGT_TAU   1
#define LGT_POS_X 2
#define LGT_POS_Y 3
#define LGT_POS_Z 4
#define LGT_POS_T 5

int frame_w=400; // used in opengl.h
int frame_h=300; // used in opengl.h
int frame=0;

#include "opengl.h"

array<int>   rndb("data/rndb.tab"); // table with random numbers
array<float> rndu("data/rndu.tab"); // table with random numbers
array<float> rndn("data/rndn.tab"); // table with random numbers

/* SINGLE RAY vs. ALL TRIANGLES
 pos
 ray
 cld result
*/
void tri_cld(float *pos,float *ray,float *cld) {
#include "tri_cld.c"
}

/* SINGLE POINT vs. ALL LIGHT SOURCES
*/
void lgt_cld(float *pnt,float *par) {
  int c1;
  float pnd[5];
  for (c1=0;c1<5;c1++)
    pnd[c1]=pnt[c1];
#define Sqrt sqrt
#include "lgt_cld.c"
}

array<float> texture("data/texture.tex");
array<float> colormap("data/colormap.rgb");
float mat[TRI_NUM*TRI_MAT]; // linear interpolation of texture coordinate
float lum[LGT_NUM*LGT_LUM];

matrix<float> obs(5,1); // coordinates of observer (never use in SHOOT function!)

void main_(int argc,char** argv) {
	obs.data[MATHEMATICA_X]=-1;
	obs.data[MATHEMATICA_Y]=3.5;
	obs.data[MATHEMATICA_Z]=5;
	obs.data[MATHEMATICA_T]=0;
  
#include "tri_mat.c" // properties of ALL TRIANGLES
#include "lgt_lum.c" // properties of ALL LIGHT SOURCES
}

/* RECURSIVE RAYTRACING FUNCTION */
void shoot(int ttl,      // time to live
           float *pos,   //    origin of ray (space-time)
           float *ray,   // direction of ray (space-time) ray.t = fraction of c
           int origin,   // triangle index in {0,1,2,...}, alternatively -1
           float factor, // weight of color (prevents recursive calls if below threshold)
           float *col    // returns color RGB
          ) {
	int c1,c2,c3,c4;
	
	/* DETERMINE TRIANGLE HIT BY RAY */
	float cld[TRI_NUM*TRI_CLD];	
	tri_cld(pos,ray,cld); // SINGLE RAY vs. ALL TRIANGLES <--------------------------------------------------
	int c1_CLD=0;         // eventually points to triangle of first collision
  float oodet,dst,tau,mu1,mu2;  
  float ref=FRUSTUM+1;  // set reference distance to maximum (since c=1 => spacial distance equals time distance)  
  int tcc;              // eventually has index of first triangle hit by ray
  for (c1=0;c1<TRI_NUM;c1++) { // for each triangle
    oodet=1/cld[c1_CLD+0];     // cld[c1_CLD+0] holds determinant of system of lin. eqs
 	  dst=cld[c1_CLD+1]*oodet;   // time for ray to reach surface (greater for slower ray)
 	  if (dst>0&&c1!=origin) {   // collision in the past and does not hit current triangle
 	    mu1=cld[c1_CLD+3]*oodet; // parameter 1 on triangle
 	    mu2=cld[c1_CLD+4]*oodet; // parameter 2 on triangle
 	    if (0<mu1&&0<mu2&&mu1+mu2<1)
   	    if (dst<ref) { // if lies in from of previous collisions...
   	      ref=dst;     // update reference distance
   	      tcc=c1;      // update index of triangle
 	      }
    }
    c1_CLD+=TRI_CLD;
  }  
  if (ref>FRUSTUM) return; // ABORT RECURSION if ray does not hit any triangle, else...

	/* INVESTIGATE WORLD FROM POINT OF COLLISION */
  c1_CLD=tcc*TRI_CLD;
  oodet=1/cld[c1_CLD+0];
  dst=cld[c1_CLD+1]*oodet; // time for ray to reach surface (greater for slower ray)
  tau=cld[c1_CLD+2]*oodet; // OBJECTS TIME !?!?! IS TAU ALWAYS NEGATIVE !?!?!
	mu1=cld[c1_CLD+3]*oodet;
	mu2=cld[c1_CLD+4]*oodet;	
	int c1_MAT,tcx,tcy;
  c1_MAT=tcc*TRI_MAT; // else point to material information of particular triangle     
  tcx=mat[c1_MAT+TEX_AFF_X]+mat[c1_MAT+TEX_MU1_X]*mu1+mat[c1_MAT+TEX_MU2_X]*mu2; // apply texture
  tcy=mat[c1_MAT+TEX_AFF_Y]+mat[c1_MAT+TEX_MU1_Y]*mu1+mat[c1_MAT+TEX_MU2_Y]*mu2; // apply texture    	  
  tcx+=tcy*texture.size[1];
  tcx*=4; // holds the index to the texture color   
  //printf("%f ",tau);
  float map=fmod(tau*mat[c1_MAT+CMAP],64); // MATLAB convention for colormap size
  if (map<0)
    map+=64;
  int cmp=map;
  cmp*=3;
  matrix<float> sug(3,1);
	sug.data[0]=texture.data[tcx++]*colormap.data[cmp++]; // suggest texture of collision point
	sug.data[1]=texture.data[tcx++]*colormap.data[cmp++]; // ...
	sug.data[2]=texture.data[tcx++]*colormap.data[cmp++]; // however, point needs to be lighted
	matrix<float> pnt(5,1);
	for (c1=0;c1<5;c1++)
    pnt.data[c1]=pos[c1]+dst*ray[c1]; // pnt is the space-time point of collision
		
	/* IF MATERIAL IS MATT CHECK FOR ILLUMINATION
	   check for direct vision to each light source
	   no recursive calls	*/
	if (mat[c1_MAT+MATT]*factor>THRESHOLD) {  	
  	float par[LGT_NUM*LGT_CLD+1];  
	  lgt_cld(pnt.data,par); // SINGLE POINT vs. ALL LIGHT SOURCES
	  // !!! MAKE SURE par HAS ABSOLUTE VALUES (not relative to pnt)
	  int c1_LGT=0,c1_LUM=0;
	  matrix<float> chk(5,1); // ray going from collision point to light source(s)
	  bool hit;
	  float lgh,prj;
	  for (c1=0;c1<LGT_NUM;c1++) { // for each light source check if anything perturbs
	    chk.data[MATHEMATICA_V]=0;
	    chk.data[MATHEMATICA_X]=par[c1_LGT+LGT_POS_X]-pnt[MATHEMATICA_X];
	    chk.data[MATHEMATICA_Y]=par[c1_LGT+LGT_POS_Y]-pnt[MATHEMATICA_Y];
	    chk.data[MATHEMATICA_Z]=par[c1_LGT+LGT_POS_Z]-pnt[MATHEMATICA_Z];
	    chk.data[MATHEMATICA_T]=0;      // transcendent
	    chk.normalize(2);
      chk.data[MATHEMATICA_T]=ray[MATHEMATICA_T]; // back to the future with the same speed as before hit
      tri_cld(pnt.data,chk.data,cld); // reuse the instance of array cld
      hit=false;
      int c2_CLD=0,c2_MAT=0;
      for (c2=0;c2<TRI_NUM;c2++) {
        oodet=1/cld[c2_CLD+0];
        lgh=pnt.data[MATHEMATICA_T]-cld[c2_CLD+1]*oodet; // absolute time of collision
        // check if time of collision is in [emmision time,collision time]
        if (par[c1_LGT+LGT_POS_T]<lgh&&lgh<pnt.data[MATHEMATICA_T]&&tcc!=c2) {
          mu1=cld[c2_CLD+3]*oodet;
          mu2=cld[c2_CLD+4]*oodet;
          if (0<mu1&&0<mu2&&mu1+mu2<1) {
            hit=true;
            break;
          }
        }
        c2_MAT+=TRI_MAT;
        c2_CLD+=TRI_CLD;
      }
      if (!hit) { // this light reaches point of triangle                          
        prj =chk.data[MATHEMATICA_X]*mat[c1_MAT+NORMAL_X];
        prj+=chk.data[MATHEMATICA_Y]*mat[c1_MAT+NORMAL_Y];
        prj+=chk.data[MATHEMATICA_Z]*mat[c1_MAT+NORMAL_Z];
        if (prj<0) prj=0; // orientation matters
        prj*=mat[c1_MAT+MATT]; // matt coefficient
        prj*=.5*(1+cos(lum[c1_LUM+COL_O]*par[c1_LGT+LGT_TAU])); // make intensity oscilaate
        col[0]+=sug.data[0]*lum[c1_LUM+COL_R]*prj;
        col[1]+=sug.data[1]*lum[c1_LUM+COL_G]*prj;
        col[2]+=sug.data[2]*lum[c1_LUM+COL_B]*prj;          
      }
      c1_LGT+=LGT_CLD;
      c1_LUM+=LGT_LUM;
    }
  }
  
  if (ttl==0) return; // ABORT RECURSION if time to live expired, else...
  
  float mat_MAT_REFL=mat[c1_MAT+REFL]; // used for TOTAL REFLECTION
  
	/* IF MATERIAL IS REFR CAST RAY IN REFLECTED DIRECTION	   
	   formula from wikipedia
	   one recursive call */	   
 	if (mat[c1_MAT+REFR]*factor>THRESHOLD) {
   	matrix<float> rfr(5,1); // holds reflected vector   	
   	float ct1,non,ct2;
    ct1 =mat[c1_MAT+NORMAL_X]*ray[MATHEMATICA_X]; // computes scalar product
    ct1+=mat[c1_MAT+NORMAL_Y]*ray[MATHEMATICA_Y]; // between normal and ray
    ct1+=mat[c1_MAT+NORMAL_Z]*ray[MATHEMATICA_Z];
    non =mat[c1_MAT+V1oV2];
    if (ct1>0) non =1/non;
    ct2 =1-non*non*(1-ct1*ct1);    
    if (ct2<0) {
      mat_MAT_REFL+=mat[c1_MAT+REFR]; // total reflection => refraction adds to strength of reflection
    } else {
      matrix<float> nuc(3,1); // holds accumulated color
      ct2 =sqrt(ct2);
      if (ct1<0) ct2*=-1;
      ct1*=-non;
      ct1+=ct2;
      rfr.data[MATHEMATICA_X]=non*ray[MATHEMATICA_X]+ct1*mat[c1_MAT+NORMAL_X]; // finalizes ray computation
      rfr.data[MATHEMATICA_Y]=non*ray[MATHEMATICA_Y]+ct1*mat[c1_MAT+NORMAL_Y];
      rfr.data[MATHEMATICA_Z]=non*ray[MATHEMATICA_Z]+ct1*mat[c1_MAT+NORMAL_Z];
      rfr.normalize(2); // CHECK IF THIS IS NECESSARY (SHOULD NOT BE)
      rfr.data[MATHEMATICA_T]=ray[MATHEMATICA_T]*non; // maybe /non instead (who knows?!??)
 	    shoot(ttl-1,                    // time to live                 
 	          pnt.data,                 //    origin of ray             
 	          rfr.data,                 // direction of refracted ray 	          
 	          tcc,                      // triangle index in {0,1,2,...}
 	          mat[c1_MAT+REFR]*factor,  // factor of color              
 	          nuc.data                  // storage of color
 	         );
      col[0]+=nuc.data[0]*mat[c1_MAT+REFR];
      col[1]+=nuc.data[1]*mat[c1_MAT+REFR];
      col[2]+=nuc.data[2]*mat[c1_MAT+REFR];    
	  }
	}
  
	/* IF MATERIAL IS REFL CAST RAY IN REFLECTED DIRECTION	   
	   one recursive call */
 	if (mat_MAT_REFL*factor>THRESHOLD) {
   	matrix<float> rfl(5,1); // holds reflected vector
   	float ryn;
    ryn =mat[c1_MAT+NORMAL_X]*ray[MATHEMATICA_X]; // computes scalar product
    ryn+=mat[c1_MAT+NORMAL_Y]*ray[MATHEMATICA_Y]; // between normal and ray
    ryn+=mat[c1_MAT+NORMAL_Z]*ray[MATHEMATICA_Z];
  	ryn*=2;
    rfl.data[MATHEMATICA_X]=ray[MATHEMATICA_X]-ryn*mat[c1_MAT+NORMAL_X]; // finalizes ray computation
    rfl.data[MATHEMATICA_Y]=ray[MATHEMATICA_Y]-ryn*mat[c1_MAT+NORMAL_Y];
    rfl.data[MATHEMATICA_Z]=ray[MATHEMATICA_Z]-ryn*mat[c1_MAT+NORMAL_Z];
    rfl.normalize(2); // CHECK IF THIS IS NECESSARY (SHOULD NOT BE)
    rfl.data[MATHEMATICA_T]=ray[MATHEMATICA_T]; // back to the future with the same speed as before hit
    matrix<float> nuc(3,1); // holds accumulated color       	
 	  shoot(ttl-1,                    // time to live                 
 	        pnt.data,                 //    origin of ray             
 	        rfl.data,                 // direction of reflected ray             
 	        tcc,                      // triangle index in {0,1,2,...}
 	        mat_MAT_REFL*factor,  // factor of color              
 	        nuc.data                  // storage of color
 	       );
    col[0]+=nuc.data[0]*mat_MAT_REFL;
    col[1]+=nuc.data[1]*mat_MAT_REFL;
    col[2]+=nuc.data[2]*mat_MAT_REFL;
	}	 
	
}

int line=0;             // current line on screen
float y;

void display_(void) {  
	int c1,cnt=0; 
	int size[3];
	float mil=2.f/frame_w;
	float x=-1;	
  line%=frame_h;
	if (line==0) { // new picture
	  y=-mil*frame_h/2;
	  ferry=obs.data[MATHEMATICA_T]; // used in OPENGL.H	  
	  if (frame) {
		  size[0]=4; size[1]=frame_w; size[2]=frame_h;
		  array<unsigned char> rgba(3,size,0);
		  glReadPixels(0,0,frame_w,frame_h,GL_RGBA,GL_UNSIGNED_BYTE,rgba.data);
		  char fnc[60];
		  sprintf(fnc,"out/pixels%i.rgba",99+frame);
		  rgba.store(fnc);
		  if (frame==FRAMES)
		    exit(0);
	  }
	  frame++;
  }
	size[0]=4; size[1]=frame_w;
	array<float> pix(2,size,(float)1.0); // holds one line of color data	
	matrix<float> ray(5,1);	
	for (c1=0;c1<pix.size[1];c1++) {
    pix.data[cnt+0]=0; // REMOVE LATER
    pix.data[cnt+1]=0; // REMOVE LATER
    pix.data[cnt+2]=0; // REMOVE LATER
    pix.data[cnt+3]=1; 
  	ray.data[MATHEMATICA_V]=0;
    ray.data[MATHEMATICA_X]=x;
    ray.data[MATHEMATICA_Y]=y;
    ray.data[MATHEMATICA_Z]=-1;
    ray.data[MATHEMATICA_T]=0;
    ray.normalize(2);           // normalize spacials
    ray.data[MATHEMATICA_T]=-1; // direct in past
    shoot(RECURSION,     // depth of recursion
          obs.data,      //    origin of ray
          ray.data,      // direction of ray
          -1,            // does not originate from any particular triangle
          (float)1,      // factor of color
          &pix.data[cnt] // storage of color
         );
  	x+=mil; // DOES THIS MAKE SENSE
  	cnt+=4;
  }
  //pix.data[0]=rndu.token()*.7; // REMOVE LATER
  //pix.data[1]=rndu.token()*.7; // REMOVE LATER
  //pix.data[2]=rndu.token()*.7; // REMOVE LATER
			
  glRasterPos2i(0,line++);
  glDrawPixels(pix.size[1],1,GL_RGBA,GL_FLOAT,pix.data);	  	
	y+=mil;
	
	if (line==frame_h)
		obs.data[MATHEMATICA_T]+=.5; // evolve observer time
}
